
from ..ioc.base import *
from buildz import dz
class ObjectType(Base):
    Single=0
    Multi=1
    SingleVars=2

class ObjectCreater(Creater):
    def init(self, obj_type = ObjectType.Single):
        super().init()
        self.obj_type = obj_type
        self.args = []
        self.maps = {}
        self.sets = {}
        self.before_set = None
        self.after_set = None
        self.before_remove = None
        self.fc = None
        self.single = Single()
    def fills(self, params):
        _a, _m, _s = dz.g(params, args=[], maps={}, sets={})
        args = self.args
        args = _a+args[len(_a):]
        maps = dict(self.maps)
        maps.update(_m)
        sets = dict(self.sets)
        sets.update(_s)
        return args, maps, sets
    def single_id(self, params):
    def hash(self, params):
        if self.obj_type == ObjectType.Single or self.obj_type == ObjectType.Multi:
            return ""
        _a, _m, _s = dz.g(params, args=[], maps={}, sets={})
        args = self.args
        args = _a+args[len(_a):]
        maps = dict(self.maps)
        maps.update(_m)
        sets = dict(self.sets)
        sets.update(_s)
        args = [str(self.mg.hash(it)) for it in args]
        maps = [str(k)+":"+self.mg.hash(val) for k,val in maps.items()]
        sets = [str(k)+":"+self.mg.hash(val) for k,val in sets.items()]
        args = ",".join(args)
        maps = ",".join(maps)
        sets = ",".join(sets)
        val = [args, maps, self.mg.hash(self.before_set), sets, self.mg.hash(self.after_set)]
        return "|".join(val)
    def call(self, params):
        call_type = dz.g(params, type="create")
        if call_type == "remove":
            self.mg.deal()
        id = None
        if self.obj_type == ObjectType.Single:
            if self.single.has(id):
                return self.single.get(id)
        _a, _m, _s = dz.g(params, args=[], maps={}, sets={})
        args = self.args
        args = _a+args[len(_a):]
        maps = dict(self.maps)
        maps.update(_m)
        sets = dict(self.sets)
        sets.update(_s)
        if self.obj_type == ObjectType.SingleVars:
            id = self.hash(params)
            if self.single.has(id):
                return self.single.get(id)
        args = [self.mg.deal(it) for it in args]
        maps = {k:self.mg.deal(v) for k,v in maps.items()}
        obj = self.fc(*args, **maps)
        params = {'target':obj}
        self.mg.deal(self.before_set, params=params)
        sets = {k:self.mg.deal(v) for k,v in sets.items()}
        for k,v in sets:
            setattr(obj, k, v)
        self.mg.deal(self.after_set, params=params)
        if self.obj_type!=ObjectType.Multi:
            self.single.set(id, obj)
        return obj


from ..base import *
basepath = path
path = pathz.Path()
path.set("conf", basepath.local("ioc/conf"), curr=0)

class Creater(Base):
    def init(self):
        self.mg = None
    def bind(self, mg):
        self.mg = mg
    def call(self, *a, **b):
        return None

pass


#
from buildz import Base

class Bind(Base):
    def init(self):
        self.mg = None
        self.binds = []
    def add_bind(self, fc):
        self.binds.append(fc)
    def bind(self, mg):
        self.mg = mg
        _ = [fc(mg) for fc in self.binds]

pass



from buildz import pyz, dz
from .tdata import TData
class Datas(TData):
    def init(self, id=None, ns=None, dts=None):
        super().init(id,ns)
        self.dts = dts
    def bind(self, dts):
        self.dts = dts
    def set(self, key, val, tag=None):
        tag = pyz.nnull(tag, self.default)
        super().set(key, val, tag)
        if self.dts is not None:
            if tag == Key.pub:
                self.dts.set(key, val, self.ns)
            elif tag == Key.ns:
                self.dts.ns_set(key, val, self.ns)
    def tget(self, key, src=None,id=None, gfind=True):
        ns, id = self.nsid(src, id)
        obj, find=super().tget(key, src, id)
        if not find and gfind:
            obj, find = self.dts.get(key, ns, id)
        return obj, find




from buildz import Base
class Args(Base):
    def str(self):
        return f"<Args args={self.args}, maps={self.maps}>"
    @property
    def lists(self):
        return self.args
    @lists.setter
    def lists(self, val):
        self.args=val
    @lists.deleter
    def lists(self):
        del self.args
    @property
    def dicts(self):
        return self.maps
    @dicts.setter
    def dicts(self, val):
        self.maps=val
    @dicts.deleter
    def dicts(self):
        del self.maps
    def init(self, args, maps):
        self.args = args
        self.maps = maps

pass


#
from .tdata import TData, TNData
class Unit(Base):
    def init(self, id=None, ns=None, deals=None):
        self.id = id
        self.ns = ns
        self.confs = Datas(id, ns)
        self.encapes = Datas(id, ns)
        self.deals = deals
        self.mg = None
    def bind(self, mg):
        self.mg = mg
        self.confs.bind(mg.confs)
        self.deals.bind(mg.deals)
    def get_deal(self, key, src=None, id=None, gfind=True):
        return self.deals.tget(key, src, id, gfind)
    def get_conf(self, key, src=None, id=None, gfind=True):
        return self.confs.tget(key, src, id, gfind)
    def get_encape(self, key, src=None, id=None, gfind=True):
        obj, find = self.encapes.tget(key, src, id)
        if not find:
            conf, cfind = self.get_conf(key, src, id, False)
            if not cfind:
                return None, 0
            
            deal, dfind = self.get_deal(key, src, id)


#
from .tdata import TData
class Deals(TData):
    def init(self, mg=None, id=None, ns=None):
        self.mg = mg
        super().init(id,ns)
    def bind(self, mg):
        self.mg = mg
    def tget(self, key, src=None,id=None):
        ns, id = self.nsid(src, id)
        obj,find=super().tget(key, src, id)
        if not find:
            obj,find = self.mg.get_deal(key, src, id)
        return obj, find


#

from .tdata import TData, TNData
from buildz import dz
class Dataset(Base):
    def init(self, ids):
        self.ids = ids
        self.pub = {}
        self.ns = {}
    def set(self, key, val, ns=None):
        keys = self.ids(ns)+self.ids(key)
        dz.dset(self.pub, keys, val)
    def ns_set(self, key, val, ns=None):
        map = dz.get_set(self.ns, ns, dict())
        map[key] = val
    def add(self, data):
        ns = data.ns
        for key, val in data.pub.items():
            self.set(key, val, ns)
        for key, val in data.ns.items()
            self.ns_set(key, val, ns)
        data.bind(self)
    def get(self, key, ns=None, id = None):
        obj, find = self.ns_get(key,ns,id)
        if not find:
            obj,find=self.pub_get(key,ns,id)
        return obj,find
    def pub_get(self, key, ns=None, id=None):
        ids = self.ids(ns)+self.ids(key)
        return dz.dget(self.pub, ids)
    def ns_get(self, key, ns=None, id=None):
        map = dz.get_set(self.ns, ns, dict())
        if key not in map:
            return None, 0
        return map[key], 1



from .base import Base, fcBase
from buildz import xf,pyz
class EnvExp(Exception):
    def __init__(self, id):
        super().__init__(f"EnvExp id not found: '{id}'")
class BuildEnv(Base):
    """
        3种从环境获取参数的方式:
            从系统环境获取
            从命令行输入参数获取
            从本对象环境数据获取
        2种设置环境参数方式：
            设置到系统环境
            设置到本对象环境数据
    """
    def init(self, conf):
        self.conf = conf
        self.builds = {
        }
    def build_args(self):
        pass
    def build_conf(self, ids, flush, flush_list, exp, default):
        return Env(ids, None, flush, flush_list, exp, default)
    def deal(self, mg, args, spt, eglobal, orders, from_args, flush, flush_list, exp, default):
        envs = []
        env = None
        ids = ids.Ids(spt)
        if 'conf' not in orders:
            orders.append('conf')
        for order in orders:
            _env = self.builds[order](ids, flush, flush_list, exp, default)
            if order == 'conf':
                env = _env
            envs.append(_env)
        env = Envs(envs, env)
        return env
    def call(self, mg, conf, args):
        args, maps = self.conf([], conf)
        return self.deal(mg, args, **maps)

pass
class Env(Base):
    def _get(self, id):
        return xf.dhget(self.maps, id)
    def _set(self, id, val):
        return xf.dset(self.maps, id, val)
    def _has(self, id):
        return xf.dhas(self.maps, id)
    def _del(self, id):
        return xf.dremove(self.maps, id)
    def _fill(self, maps):
        return xf.fill(maps, self.maps)
    def init(self, ids=None, maps=None, flush=True, flush_list = False):
        self.flush_list = flush_list
        self.ids = ids
        self.flush = flush
        if maps is None:
            maps = {}
        if flush:
            maps=xf.flush_maps(maps, self.ids.ids, flush_list)
        self.maps = maps
    def update(self, maps, flush=None, flush_list=None):
        flush = pyz.nnull(flush, self.flush)
        flush_list = pyz.nnull(flush_list, self.flush_list)
        if flush:
            maps = xf.flush_maps(maps, self.ids.ids, flush_list)
        self._fill(maps)
    def has(self, id, flush=None):
        flush = pyz.not_null(flush, self.flush)
        full_id = id
        if flush:
            ids = self.ids.ids(id)
        else:
            ids = id
        return self._has(ids)
    def get(self, id, flush=None):
        flush = pyz.not_null(flush, self.flush)
        full_id = id
        if flush:
            ids = self.ids.ids(id)
        else:
            ids = id
        return self._get(ids)
    def set(self, id, val, flush=None):
        flush = pyz.not_null(flush, self.flush)
        if flush:
            id = self.ids.ids(id)
        self._set(self.maps, id, val)
 
class SysEnv(Env):
    def _get(self, id, exp_def=None):
        sysdt = os.getenv(id)
        return sysdt
    def _has(self, id):
        dt = self._get(id)
        return dt is not None
    def init(self, ids):
        super().init(ids, flush=False)

class Envs(Env):
    def init(self, envs, env):
        self.envs = envs
        self.env = env
    def get(self, id):
        for env in self.envs:
            if env.has(id):
                return env.get(id)
        return self.env.get(id)
    def set(self, id, val, *a,**b):
        return self.env.set(id, val, *a,**b)
    def has(self, id):
        for env in self.envs:
            if env.has(id):
                return True
        return self.env.has(id)
    def update(self, maps, *a,**b):
        return self.env.update(maps,*a,**b)



#
from .tdata import TData
from .datas import Datas
from .dataset import Dataset
class Encpaes(Datas):
    def init(self, confs, deals, id=None, ns=None, dts=None):
        super().init(id, ns, dts)
    def tget(self, key, src=None,id=None, gfind=True):
class Encapeset(Dataset):
    '''
        生成对象前的中间层
    '''
    def init(self, ids, confs, deals):
        super().init(ids)
        self.confs = confs
        self.deals = deals
    def pub_get(self, key, ns=None, id=None):
        obj, find = super().get(key, ns, id)

    def ns_get(self, key, ns=None, deal_ns=None):
        map = dz.get_set(self.ns, ns, dict())
        if key not in map:
            return None, 0
        return map[key], 1

    